#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#define NDIRS   2 /* lazily static */

struct events {
    int flag;
    char *name;
} events[] = {
    { DN_ACCESS, "accessed"},
    { DN_MODIFY, "modified"},
    { DN_CREATE, "created"},
    { DN_DELETE, "deleted"},
    { DN_RENAME, "renamed"},
    { DN_ATTRIB, "attr-changed"}
};

#define NEVENTS (sizeof(events)/sizeof(struct events))

struct direvents {
    int fd[NEVENTS];
    int count[NEVENTS];
} direvents[NDIRS];

unsigned int sigcount;

void handler(int signo, siginfo_t *info, void *v)
{
    int i, j;

    /* we only get this signal, so no check */
    for (i = 0; i < NDIRS; i++) {
	for (j = 0; j < NEVENTS; j++) {
	    if (info->si_fd == direvents[i].fd[j])
		direvents[i].count[j]++;
	}
    }
    sigcount++;
}

int main(int argc, char **argv)
{
    struct sigaction sa;
    char **dirs;
    int i, j;

    if (argc != NDIRS+1) {
	fprintf(stderr, "%s: Use: \"%s <dir> <dir> ...\" (%i dirs)\n",
		argv[0], argv[0], NDIRS);
	exit(1);
    }
    dirs = argv+1;

    /* prepare signal handler for SIGUSR1 */
    memset(&sa, 0, sizeof(sa));
    sa.sa_sigaction = handler;
    sa.sa_flags = SA_SIGINFO;
    if (sigaction(SIGUSR1, &sa, NULL)) {
	fprintf(stderr, "%s: sigaction(): %s\n", argv[0], strerror(errno));
	exit(1);
    }

    /* set up stuff, this also checks for errors */
    for (i = 0; i < NDIRS; i++) {
	for (j = 0; j < NEVENTS; j++) {
	    int fd = open(dirs[i], O_RDONLY);

	    if (fd < 0) {
		fprintf(stderr, "%s: %s: %s\n", argv[0], dirs[i],
			strerror(errno));
		exit(1);
	    }
	    direvents[i].fd[j] = fd;

	    if (fcntl(fd, F_SETSIG, SIGUSR1)) {
		fprintf(stderr, "%s: %s: F_SETSIG: %s\n", argv[0], dirs[i],
			strerror(errno));
		exit(1);
	    }
	    if (fcntl(fd, F_NOTIFY, events[j].flag | DN_MULTISHOT)) {
		fprintf(stderr, "%s: %s: F_NOTIFY: %s\n", argv[0], dirs[i],
			strerror(errno));
		exit(1);
	    }
	}
    }

    while (1) {
	if (!sigcount) select(0, NULL, NULL, NULL, NULL); /* forever */
	sigcount = 0;
	for (i = 0; i < NDIRS; i++) {
	    for (j = 0; j < NEVENTS; j++) {
		if (direvents[i].count[j]) {
		    direvents[i].count[j] = 0;
		    printf("in %s: file %s\n", dirs[i], events[j].name);
		}
	    }
	}
    }
    exit(0);
}
